// Copyright (c) 2020 GeometryFactory Sarl (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.2/Arrangement_on_surface_2/demo/Arrangement_on_surface_2/QtMetaTypes.h $
// $Id: demo/Arrangement_on_surface_2/QtMetaTypes.h cac3e9d75e2 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s): Ahmed Essam <theartful.ae@gmail.com>

#ifndef ARRANGEMENT_DEMO_QT_METATYPES_H
#define ARRANGEMENT_DEMO_QT_METATYPES_H

#include <QMetaType>
#include <CGAL/Object.h>

Q_DECLARE_METATYPE(CGAL::Object)

#endif
