// Copyright (c) 2023  GeometryFactory (France).  All rights reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.2/BGL/include/CGAL/boost/graph/IO/polygon_mesh_io.h $
// $Id: include/CGAL/boost/graph/IO/polygon_mesh_io.h cac3e9d75e2 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s)     : Mael Rouxel-Labbé

#ifndef CGAL_BOOST_GRAPH_POLYGON_MESH_IO_H
#define CGAL_BOOST_GRAPH_POLYGON_MESH_IO_H


#include <CGAL/IO/polygon_mesh_io.h>


#endif // CGAL_BOOST_GRAPH_POLYGON_MESH_IO_H
