// Copyright (c) 2012  GeometryFactory Sarl (France)
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.2/Mesh_3/include/CGAL/Mesh_3/Dump_c3t3.h $
// $Id: include/CGAL/Mesh_3/Dump_c3t3.h cac3e9d75e2 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Laurent Rineau

#ifndef CGAL_MESH_3_DUMP_C3T3_H
#define CGAL_MESH_3_DUMP_C3T3_H

#include <CGAL/license/SMDS_3.h>

#include <CGAL/SMDS_3/Dump_c3t3.h>

#endif // CGAL_MESH_3_DUMP_C3T3_H
