// Copyright (c) 2023  GeometryFactory Sarl (France)
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL$
// $Id$
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s)     : Laurent Rineau

#include <CGAL/Three/Scene_item_with_properties.h>

CGAL::Three::Scene_item_with_properties::~Scene_item_with_properties() {}
