// Copyright (c) 2009 INRIA Sophia-Antipolis (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.2/Mesh_3/include/CGAL/Labeled_mesh_domain_3.h $
// $Id: include/CGAL/Labeled_mesh_domain_3.h cac3e9d75e2 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s)     : Stéphane Tayeb, Aymeric PELLE


#ifndef CGAL_LABELED_MESH_DOMAIN_3_H
#define CGAL_LABELED_MESH_DOMAIN_3_H

#include <CGAL/license/Mesh_3.h>

#include <CGAL/Mesh_3/Labeled_mesh_domain_3_implicit_function.h>
#include <CGAL/Mesh_3/Labeled_mesh_domain_3_image.h>

#endif // CGAL_LABELED_MESH_DOMAIN_3_H
