// Copyright (c) 2019-2022 Google LLC (USA).
// Copyright (c) 2025 GeometryFactory (France)
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.2/Alpha_wrap_2/include/CGAL/Alpha_wrap_2/internal/oracles.h $
// $Id: include/CGAL/Alpha_wrap_2/internal/oracles.h cac3e9d75e2 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s)     : Mael Rouxel-Labbé
//
#ifndef CGAL_ALPHA_WRAP_2_INTERNAL_ORACLES_H
#define CGAL_ALPHA_WRAP_2_INTERNAL_ORACLES_H

#include <CGAL/license/Alpha_wrap_2.h>

#include <CGAL/Alpha_wrap_2/internal/Alpha_wrap_AABB_geom_traits.h>
#include <CGAL/Alpha_wrap_2/internal/offset_intersection.h>

#include <CGAL/Alpha_wrap_2/internal/Point_set_oracle.h>
#include <CGAL/Alpha_wrap_2/internal/Segment_soup_oracle.h>

#endif // CGAL_ALPHA_WRAP_2_INTERNAL_ORACLES_H
